unit modMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls;

type
  TfrmMain = class(TForm)
    GroupBox1: TGroupBox;
    edtSource: TEdit;
    btnSource: TButton;
    btnTarget: TButton;
    edtTarget: TEdit;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    btnConvert: TButton;
    btnExit: TButton;
    lblStatus: TLabel;
    edtBlockSize: TEdit;
    lblBlock1: TLabel;
    lblBlock2: TLabel;
    lblBlockSize: TLabel;
    procedure btnSourceClick(Sender: TObject);
    procedure btnTargetClick(Sender: TObject);
    procedure btnConvertClick(Sender: TObject);
    procedure btnExitClick(Sender: TObject);
    procedure edtTargetChange(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure edtBlockSizeChange(Sender: TObject);
  private
    { Private-Deklarationen }
    bCancel: Boolean;
  public
    { Public-Deklarationen }
  end;

var
  frmMain: TfrmMain;

implementation

{$R *.DFM}

const
  DEF_BLOCK_SIZE = 1048576; // 1 MB

var
  sTargetDir : String;
  iBlockSize : Int64;

procedure TfrmMain.btnSourceClick(Sender: TObject);
begin
  OpenDialog1.FileName := edtSource.Text;
  if (OpenDialog1.Execute) then
    begin
      edtSource.Text := OpenDialog1.FileName;
      if sTargetDir = '' then
        edtTarget.Text := ChangeFileExt(edtSource.Text, '.iso')
      else
        edtTarget.Text := ChangeFileExt(sTargetDir + ExtractFileName(edtSource.Text), '.iso');
    end;
end;

procedure TfrmMain.btnTargetClick(Sender: TObject);
begin
  SaveDialog1.FileName := edtTarget.Text;
  if (SaveDialog1.Execute) then
    edtTarget.Text := SaveDialog1.FileName;
end;

// (coded) Abstract the set file pointer routine in WinNT so we don't 
// have to manually do the comp arithmetic and manipulation of 
// pointers.  
function SetFilePointerComp(
		hFile : THandle; 
		const seekPos : comp;
		moveMethod : cardinal) : cardinal;
type
   PCardinal = ^cardinal;         
var
   lowFPosVal : cardinal;
   highFPosPtr : PCardinal;
begin
   // Get the low-order long via a pointer.
   highFPosPtr := @seekPos;
   lowFPosVal := highFPosPtr^;

   // Point the pointer to the high order value.
   highFPosPtr := pointer(cardinal(highFPosPtr) + 4);
   result := SetFilePointer(hFile, lowFPosVal, highFPosPtr,
                moveMethod);
end;

{$O-}
// (debugged) A procedure to get a comp level file size and handle all
// the file arithmetic for for you.  You simply get a comp back.
function GetFileSizeComp(hFile : THandle) : Int64;
type
   PCardinal = ^cardinal;         
var
   tComp : Int64;
   tLong, tlong1 : cardinal;
   highFPosPtr : PCardinal;
begin
   // Get the low-order long via a pointer.
   tComp := 0;

   // Remember, -1 return is not
   tLong := GetFileSize(hFile, @tlong1);
   if (tLong = $ffffffff) then 
      result := -1
   else begin
      // We have to do these stupid moves because Borland Delphi does 
      // not support a real cardinal, it supports a longint max on 
      // cardinal, so we get negative numbers between 2 billion and 4 
      // billion file size.
      highFPosPtr := pointer(cardinal(@tComp) + 4);
      Move(tLong, tComp, SizeOf(tLong));
      Move(tLong1, highFPosPtr^, Sizeof(tLong1));
      result := tcomp;
   end;      
end;         
{$O+}

function ConvertFile : Integer;
var
  src, dst: Int64;
  buf: Pointer;
  n : LongInt;
  bytestocopy : Int64;
  totalbytes  : Int64;
  bytes : Int64;
  lastpart: array [0..4] of Int64;
begin
  with frmMain do
    begin
      Tag := 1;
      bCancel := False;
      btnSource.Enabled := False;
      btnTarget.Enabled := False;
      btnExit.Enabled := False;
      edtBlockSize.Enabled := False;
      btnConvert.Caption := '&Abort';
      src := INVALID_HANDLE_VALUE;
      dst := INVALID_HANDLE_VALUE;
      try
        GetMem(buf, (iBlockSize*DEF_BLOCK_SIZE));
        src := CreateFile(PChar(edtSource.Text), GENERIC_READ, FILE_SHARE_READ, nil, OPEN_EXISTING, 0, 0);
        if (src = INVALID_HANDLE_VALUE) then
          raise EInOutError.Create('File not found');
        dst := CreateFile(PChar(edtTarget.Text), GENERIC_WRITE, FILE_SHARE_READ, nil, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
        if (dst = INVALID_HANDLE_VALUE) then
          raise EInOutError.Create('Error creating file');
        bytestocopy := GetFileSizeComp(src);
        totalbytes := bytestocopy;
        //bytestocopy := SetFilePointerComp(src, 0, FILE_END);
        SetFilePointerComp(src, -sizeof(lastpart), FILE_END);
        if (not ReadFile(src, lastpart, sizeof(lastpart), DWord(n), nil)) then
          raise EInOutError.Create('Error, File too short');
        if (((lastpart[0] = $21444E45) and (lastpart[2] = $3552454E)) or
            ((lastpart[1] = $21444E45) and (lastpart[3] = $4F52454E))) then
          begin
            bytestocopy := lastpart[4];
            asm
              mov eax, Longint(bytestocopy)
              xchg ah, al
              rol eax, 16
              xchg ah, al
              mov LongInt(bytestocopy), eax
            end;
          end;
        Dec(bytestocopy, 512*600);
        if (SetFilePointerComp(src, 512*600, FILE_BEGIN) = -1) then
          raise EInOutError.Create('Error, File too short');
        while (bytestocopy > 0) do
          begin
            bytes := bytestocopy;
            if (bytes > (iBlockSize*DEF_BLOCK_SIZE)) then
              bytes := (iBlockSize*DEF_BLOCK_SIZE);
            if (not ReadFile(src, buf^, bytes, DWord(n), nil)) then
              raise EInOutError.Create('Error Reading file');
            if n = 0 then
              break;
            Dec(bytestocopy, n);
            if bCancel then
              break;
            if (not WriteFile(dst, buf^, n, DWord(n), nil)) then
              raise EInOutError.Create('Error Writing file');
            //lblStatus.Caption := Format('%d Bytes copied', [SetFilePointerComp(src, 0, FILE_CURRENT)]);
            lblStatus.Caption := Format('%u Bytes copied', [totalbytes-bytestocopy]);
            lblStatus.Update;
            Application.ProcessMessages;
          end;
        if bCancel then
          lblStatus.Caption := 'Convert Cancelled'
        else
          lblStatus.Caption := 'Finished';
      except
        on e: EInOutError do
          lblStatus.Caption := 'Error: ' + e.Message;
      end;
      CloseHandle(src);
      CloseHandle(dst);
      btnConvert.Caption := '&Convert';
      btnSource.Enabled := True;
      btnTarget.Enabled := True;
      btnExit.Enabled := True;
      edtBlockSize.Enabled := True;
      try
        if buf <> nil then
          FreeMem(buf);
      except
      end;
      Tag := 0;
    end;

  EndThread(0);
end;

procedure TfrmMain.btnConvertClick(Sender: TObject);
var
  ThreadID : DWORD;
begin
  if Tag = 0 then
  begin
    try
      BeginThread(nil, 0, @ConvertFile, Pointer(Self), 0, ThreadID);
    except
      On E:Exception do
      begin
         lblStatus.Caption := 'Exception launching Convert Thread : '+E.Message;
      end;
    end;
  end
  else
    bCancel := True;
end;


procedure TfrmMain.btnExitClick(Sender: TObject);
begin
  Close;
end;

procedure TfrmMain.edtTargetChange(Sender: TObject);
var
  sWork : String;
begin
  sWork := ExtractFilePath(edtTarget.Text);
  if sWork <> '' then
    sTargetDir := sWork;
end;

procedure TfrmMain.FormCreate(Sender: TObject);
begin
  iBlockSize := 1;
  edtBlockSize.Text := IntToStr(iBlockSize);
end;

{$HINTS OFF}
function IsNumeric(const AString: string): Boolean;
var
  LCode: Integer;
  LVoid: Integer;
begin
  Val(AString, LVoid, LCode);
  Result := LCode = 0;
end;
{$HINTS ON}

procedure TfrmMain.edtBlockSizeChange(Sender: TObject);
begin
  if IsNumeric(edtBlockSize.Text) then
  begin
    iBlockSize := StrToInt(edtBlockSize.Text);
    lblBlockSize.Caption := IntToStr(iBlockSize*DEF_BLOCK_SIZE);
  end;
end;

end.
